/* This file is part of GNU RADIUS.
   Copyright (C) 2000,2001 Sergey Poznyakoff
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>

#ifdef HAVE_TERMCAP_H
# include <termcap.h>
#endif
#include <sys/ioctl.h>

#include <slogin.h>

static	int	sig_caught;
#ifdef HAVE_SIGACTION
static	struct	sigaction sigact;
#endif

/*ARGSUSED*/
static RETSIGTYPE
sig_catch(int sig)
{
	sig_caught = 1;
#ifdef USE_SETJMP
	siglongjmp(intr, 1);
#endif
}


#if defined(HAVE_TERMIO_H)
#include <termio.h>

struct termio old_settings, new_settings;
# define save_settings() ioctl(1, TCGETA, &old_settings) 

int
sysdep_init_screen()
{
	if (ioctl(1, TCGETA, &old_settings) != -1) {
		new_settings = old_settings;
		new_settings.c_lflag &= ~(ECHO|ECHOE|ECHOK);
		ioctl(1, TCSETA, &new_settings);

		return 0;
	}
	return 1;
}

int
sysdep_restore_screen()
{
	ioctl(1, TCSETA, &old_settings);
}

#elif defined(HAVE_TERMIOS_H)
#include <termios.h>

struct termios old_settings, new_settings;
# define save_settings() tcgetattr(1, &old_settings)

int
sysdep_init_screen()
{
	if (tcgetattr(1, &old_settings) != -1) {
		new_settings = old_settings;
		new_settings.c_lflag &= ~(ECHO);
		tcsetattr(1, TCSADRAIN, &new_settings);
		return 0;
	}
	return 1;
}

int
sysdep_restore_screen()
{
	tcsetattr(1, TCSADRAIN, &old_settings);
}

#elif defined(HAVE_SGTTY_H)
#include <sgtty.h>
struct sgttyb old_settings, new_settings;
int old_lword, new_lword;
# define save_settings() ioctl(1, TIOCGETP, &old_settings) 

int
sysdep_init_screen()
{
	if (ioctl(1, TIOCGETP, &old_settings) != -1) {
		/*
		 * cbreak on,
		 * echo off
		 * tab expansion off
		 */
		new_settings = old_settings;
		new_settings.sg_flags |= CBREAK;
		new_settings.sg_flags &=  ~(ECHO|ECHOE|ECHOK);
		ioctl(1, TIOCSETP, &new_settings);

#ifdef TOStop
		/* preserve and modify local mode word */
		ioctl(1, TIOCLGET, &old_lword);
		new_lword = old_lword | LTOSTOP;
		ioctl(1, TIOCLSET, &new_lword);
#endif
		return 0;
	}
	return 1;
}

int
sysdep_restore_screen()
{
	ioctl(1, TIOCSETP, &old_settings);
#ifdef TOStop
	ioctl(1, TIOCLSET, &old_lword);
#endif
}
#else
# error "You loose: none of sgtty.h termio.h termios.h found!"
#endif

static char *
slogin_read(FILE *fp)
{
	static char input[MAXLEN + 1];
	char *cp;
	
	fflush(stdout);
	if (fgets(input, sizeof input, fp) == input) {
		if ((cp = strchr (input, '\n')))
			*cp = '\0';
		else
			input[sizeof input - 1] = '\0';
#ifdef USE_SGTTY
		putc('\n', stdout);
#endif
		return input;
	}
	return NULL;
}

char *
slogin_getuser(const char *prompt)
{
	if (fputs(prompt, stdout) == EOF)
		abort();
	
	return slogin_read(stdin);
}

char *
slogin_getpass(const char *prompt)
{
	static char *return_value;
	static int tty_opened;
	static FILE *fp;
#ifdef HAVE_SIGACTION
	struct sigaction old_sigact;
#else
	RETSIGTYPE (*old_signal)();
#endif

	/*
	 * set a flag so the SIGINT signal can be re-sent if it
	 * is caught
	 */

	sig_caught = 0;
	return_value = 0;
	tty_opened = 0;

	/*
	 * if /dev/tty can't be opened, getpass() needs to read
	 * from stdin instead.
	 */
#if 0
	if ((fp = fopen ("/dev/tty", "r")) == 0) {
		fp = stdin;
		setbuf (fp, (char *) 0);
	} else {
		tty_opened = 1;
	}
#else
	fp = stdin;
#endif
	
	if (sysdep_init_screen()) 
		abort();


#ifdef USE_SETJMP
	/*
	 * If we get a SIGINT, sig_catch() will jump here -
	 * no need to press Enter after Ctrl-C.
	 */
	if (sigsetjmp(intr, 1))
		goto out;
#endif

#ifdef HAVE_SIGACTION
	sigact.sa_handler = sig_catch;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	sigaction(SIGINT, &sigact, &old_sigact);
#else
	old_signal = signal (SIGINT, sig_catch);
#endif

	/*
	 * the prompt is output, and the response read without
	 * echoing.  the trailing newline must be removed.  if
	 * the fgets() returns an error, a NULL pointer is
	 * returned.
	 */

	if (fputs(prompt, stdout) == EOF)
		goto out;

	return_value = slogin_read(fp);
out:
	/*
	 * the old SIGINT handler is restored after the tty
	 * modes.  then /dev/tty is closed if it was opened in
	 * the beginning.  finally, if a signal was caught it
	 * is sent to this process for normal processing.
	 */

	sysdep_restore_screen();

#ifdef HAVE_SIGACTION
	(void) sigaction (SIGINT, &old_sigact, NULL);
#else
	(void) signal (SIGINT, old_signal);
#endif

	if (tty_opened)
		fclose (fp);

	if (sig_caught) {
		kill(getpid(), SIGINT);
		return_value = NULL;
	}

	putchar('\n');
	fflush(stdout);
	return return_value;
}
